unit uJclExceptInfo;

{******************************************************************************
Initial Author   :    Scott Price
Initially Created:    09-Aug-2002
Language         :    Object Pascal
Compiler Version :    Delphi 3.02, 4.03, 5.01, 6.02 (UK)
Copyright        :     Information Technology Sector Limited, 1996-2002.
                      Sections of code taken from Jedi Code Library example
                      and subject to the Project JEDI Code Library (JCL)
                      Software Licence.  Code elements taken from the file:

  [Jcl Install Dir]\examples\debugextension\dialog\ExceptDlg.pas

Source Code Constraints:
========================
   Distribution, disclosure or modifications of this source code, outside of
   any licenced agreements with Information Technology Sector Limited, or
   removal of this notice constitutes a breach of any or all license agreements.

Operating Systems Tested On:
============================
   Microsoft Windows(R) 98, Second Edition
   Microsoft Windows(R) 2000, Service Pack 2

Design Notes:
=============
  To provide a non-visual encapsulation of the Exception information that can
  be obtained through the use of the JclDebug unit routines, into a form that
  will allow us to be notified via an event that we have received any new
  information for what-ever we want to do with it.

VERSION  AUTHOR                DATE/TIME
===============================================================================
$history$
$nokeywords$
===============================================================================
NOTES:
 - BE WARNED!!!  I Haven't tested this object in a multi-threaded environment,
   and in its current state it COULD be a PROBLEM to use in that environment,
   dependant on how you use it.  However, it examples how to use the
   JclDebug/JclHookExcept units routines.  If you examine the routine called:

     JclHookExcept.JclAddExceptNotifier
     
   You will see that there are two overloaded version.  One takes a method of
   an object as the means to report back issues, and the other takes a
   procedure as a common means to report to.  Your call as to how you would
   prefer to work.  But consider yourselves warned  ;-)
===============================================================================
TODO:
 1) For the future, might be nice to provide a means to modify the Tracking
    Options as a property of TCustomExceptionInformation, for more control
    of all of the available options
******************************************************************************}

{**************************************************************************************************}
{                                                                                                  }
{ Project JEDI Code Library (JCL)                                                                  }
{                                                                                                  }
{ The contents of this file are subject to the Mozilla Public License Version 1.1 (the "License"); }
{ you may not use this file except in compliance with the License. You may obtain a copy of the    }
{ License at http://www.mozilla.org/MPL/                                                           }
{                                                                                                  }
{ Software distributed under the License is distributed on an "AS IS" basis, WITHOUT WARRANTY OF   }
{ ANY KIND, either express or implied. See the License for the specific language governing rights  }
{ and limitations under the License.                                                               }
{                                                                                                  }
{ The Original Code is ExceptDlg.pas.                                                              }
{                                                                                                  }
{ The Initial Developer of the Original Code is documented in the accompanying                     }
{ help file JCL.chm. Portions created by these individuals are Copyright (C) of these individuals. }
{                                                                                                  }
{**************************************************************************************************}
{                                                                                                  }
{ Sample exception information capture for non-visual Applications                                 }
{                                                                                                  }
{ Last modified: September 23, 2002                                                                }
{                                                                                                  }
{**************************************************************************************************}


interface


uses
  { Jedi Code Library Units }
  JclHookExcept,

  { Borland Standard Units }
  Classes;


type
  { TCustomExceptionInformation }

  TExceptionNotifyEvent = procedure(Sender: TObject; const Info: TStrings) of object;

  TCustomExceptionInformation = class(TPersistent)
  private
    { Private Declarations }
    FHooked: Boolean;
    FExceptionInfo: TStrings;
    FOnException: TExceptionNotifyEvent;
    FRawInformation: Boolean;
    procedure SetRawInformation(const Value: Boolean);
    procedure SetTrackingOptions;
    procedure SetExceptionInfo(const Value: TStrings);
  protected
    { Protected Declarations }
    procedure ExceptionNotifier(ExceptObj: TObject; ExceptAddr: Pointer;
      OSException: Boolean);
    procedure CreateExceptionInfo(ExceptObj: TObject;
      ExceptAddr: Pointer; OSException: Boolean); virtual; abstract;
    procedure DoOnException;
    property Hooked: Boolean read FHooked;
    property ExceptionInfo: TStrings read FExceptionInfo write SetExceptionInfo;
  public
    { Public Declarations }
    constructor Create; virtual;
    constructor CreateHooked(const EventHandler: TExceptionNotifyEvent); virtual;
    destructor Destroy; override;
    procedure HookExceptions;
    procedure UnHookExceptions;
  published
    { Published Declarations }
    property RawInformation: Boolean read FRawInformation write SetRawInformation;
    property OnException: TExceptionNotifyEvent read FOnException write FOnException;
  end;


  { TStandardExceptionInfo }

  TStandardExceptionInfo = class(TCustomExceptionInformation)
  private
    { Private Declarations }
    FWriteSeperator: Boolean;
    function GetBPP: Integer;
    function GetDisplayWidth: Integer;
    function GetDisplayHeight: Integer;
    procedure SetWriteSeperator(const Value: Boolean);
    { The methods responsible for specific segments of the useful information }
    procedure GetLoadedModulesList;
    procedure GetJclDebugStackList;
    procedure GetOSInformation;
  protected
    { Protected Declarations }
    procedure CreateExceptionInfo(ExceptObj: TObject;
      ExceptAddr: Pointer; OSException: Boolean); override;
  public
    { Public Declarations }
    constructor Create; override;
  published
    { Published Declarations }
    { Change Visibility of these properties to Published from Protected }
    property ExceptionInfo;
    property Hooked;
    property WriteSeperator: Boolean read FWriteSeperator write SetWriteSeperator;
  end;



implementation



uses
  { Jedi Code Library Units }
  JclDebug, JclSysInfo, JclStrings, JclFileUtils,

  { Borland Standard Units }
  SysUtils, Windows;


  { TCustomExceptionInformation }

constructor TCustomExceptionInformation.Create;
begin
  inherited Create;

  FHooked := False;
  FExceptionInfo := TStringList.Create;

  { Default to the Raw Information - More Details help }
  FRawInformation := True;

  FOnException := Nil;

  SetTrackingOptions;
end;

constructor TCustomExceptionInformation.CreateHooked(
  const EventHandler: TExceptionNotifyEvent);
begin
  { Do ths standard create, and also set-up the Event straight away }
  Create;

  Self.FOnException := EventHandler;
  Self.HookExceptions;
end;

destructor TCustomExceptionInformation.Destroy;
begin
  TRY
    { Try to release any calls made to the HookExceptions if we are still
      Hooked to the exception notification queues/list }
    if FHooked then
      UnHookExceptions;

    if Assigned(FOnException) then
      FOnException := Nil;

    if Assigned(FExceptionInfo) then
      FreeAndNil(FExceptionInfo);
  FINALLY
    inherited Destroy;
  END;
end;

procedure TCustomExceptionInformation.DoOnException;
begin
  if Assigned(FOnException) then
    FOnException(Self, FExceptionInfo);
end;

procedure TCustomExceptionInformation.ExceptionNotifier(ExceptObj: TObject;
  ExceptAddr: Pointer; OSException: Boolean);
begin
  { TODO:  React to the exception information }
  CreateExceptionInfo(ExceptObj, ExceptAddr, OSException);

  DoOnException;
end;

procedure TCustomExceptionInformation.HookExceptions;
begin
  { Start Exception tracking }
  if NOT JclDebug.JclExceptionTrackingActive then
    JclDebug.JclStartExceptionTracking;

  { Assign notification procedure for hooked RaiseException API call. This
    allows being notified of any exception }
  JclHookExcept.JclAddExceptNotifier(ExceptionNotifier);

  FHooked := True;
end;

procedure TCustomExceptionInformation.SetExceptionInfo(
  const Value: TStrings);
begin
  FExceptionInfo.Assign(Value);
end;

procedure TCustomExceptionInformation.SetRawInformation(
  const Value: Boolean);
begin
  if (FRawInformation <> Value) then
  begin
    FRawInformation := Value;

    { This configuration line determines the amount of information returned to
      you by the ExceptionNotifier, and can be a lot more if in stRawMode.  To
      see the difference, un-comment the first line and comment out the second
      line, etc...  However, without this full Raw Information you might not
      be provided with information about Your Unit that might have invoked the
      calls to the appropriate items, but rather just those last few leading
      up to the actual exception itself.  The best way to see the difference is
      to compile and run both options in seperate compiles }

    SetTrackingOptions;
  end;
end;

procedure TCustomExceptionInformation.SetTrackingOptions;
begin
  { Set-up the appropriate set items to determine the amount of information
    we wish to receive from the Stack List when we get an exception } 
  if FRawInformation then
    Include(JclDebug.JclStackTrackingOptions, stRawMode)
  else
    Exclude(JclDebug.JclStackTrackingOptions, stRawMode);
end;

procedure TCustomExceptionInformation.UnHookExceptions;
begin
  { Try to remove our exception notifier }
  if FHooked then
    TRY
      JclHookExcept.JclRemoveExceptNotifier(ExceptionNotifier);
    FINALLY
      FHooked := False;
    END;
end;


  { TStandardExceptionInfo }

resourcestring
  RsExceptionClass = 'Exception Class: %s';
  RsExceptionAddr = 'Exception Address: %p';
  RsExceptionMessage = 'Exception Message: %s';
  RsStackList = 'Stack list, generated %s';
  RsModulesList = 'List of loaded modules:';
  RsOSVersion = 'OS: Win%s, Version: %d.%d, Build: %x, "%s"';
  RsProcessor = 'Processor: %s %s %d %s%s';
  RsScreenRes = 'Screen: %dx%d pixels, %d bpp';
  RsMainThread = 'IsMainThread: %s';
  RsThreadID = 'Thread ID: %d';


constructor TStandardExceptionInfo.Create;
begin
  inherited Create;

  FWriteSeperator := False;
end;

function TStandardExceptionInfo.GetBPP: Integer;
var
  DC: HDC;
begin
  { Routine to get the Current Screen Colour Depth }
  DC := GetDC(0);
  Result := GetDeviceCaps(DC, BITSPIXEL) * GetDeviceCaps(DC, PLANES);
  ReleaseDC(0, DC);
end;

function TStandardExceptionInfo.GetDisplayHeight: Integer;
begin
  Result := GetSystemMetrics(SM_CYSCREEN);
end;

function TStandardExceptionInfo.GetDisplayWidth: Integer;
begin
  Result := GetSystemMetrics(SM_CXSCREEN);
end;

procedure TStandardExceptionInfo.CreateExceptionInfo(ExceptObj: TObject;
  ExceptAddr: Pointer; OSException: Boolean);
var
  bIsMainThread: Boolean;
begin
  { Abstract Method lies below - do not call Inherited }
  ExceptionInfo.BeginUpdate;
  try
    { Basic Exception Information }
    bIsMainThread := (GetCurrentThreadId = MainThreadID);

    if FWriteSeperator then
      ExceptionInfo.Add('============================================================');

    ExceptionInfo.Add(Format(RsExceptionClass, [ExceptObj.ClassName]));
    ExceptionInfo.Add(Format(RsExceptionAddr, [ExceptAddr]));
    if ExceptObj is Exception then
      ExceptionInfo.Add(Format(RsExceptionMessage, [Exception(ExceptObj).Message]));
    ExceptionInfo.Add(Format(RsMainThread, [JclStrings.BooleanToStr(bIsMainThread)]));
    ExceptionInfo.Add(Format(RsThreadID, [GetCurrentThreadId]));
    ExceptionInfo.Add('');

    { Get the JclDebug Stack List, some OS information, and the Loaded Modules
      list.  These items are frequently requested pieces of information for
      some developers trying to trouble-shoot system problems. }
    GetJclDebugStackList;

    GetOSInformation;

    GetLoadedModulesList;
  finally
    ExceptionInfo.EndUpdate;
  end;
end;

procedure TStandardExceptionInfo.SetWriteSeperator(const Value: Boolean);
begin
  if FWriteSeperator <> Value then
    FWriteSeperator := Value;
end;

procedure TStandardExceptionInfo.GetLoadedModulesList;
var
  slModules: TStringList;
  I: Integer;
  sModuleName: String;
begin
  { Try to Retrieve the Loaded Modules list }
  slModules := TStringList.Create;
  TRY
    if LoadedModulesList(slModules, GetCurrentProcessId) then
    begin
      ExceptionInfo.Add(RsModulesList);
      slModules.Sort;
      for I := 0 to (slModules.Count - 1) do
      begin
        sModuleName := slModules[I];
        ExceptionInfo.Add(Format('[%.8x] %s', [DWORD(slModules.Objects[I]), sModuleName]));

        { Try to obtain some basic file version information for each item also }
        if JclFileUtils.VersionResourceAvailable(sModuleName) then
          with TJclFileVersionInfo.Create(sModuleName) do
            try
              ExceptionInfo.Add(JclStrings.StrRepeat(' ', 11) + BinFileVersion + ' - ' + FileVersion);
              if FileDescription <> '' then
                ExceptionInfo.Add(StrRepeat(' ', 11) + FileDescription);
            finally
              Free;
            end;
      end;
    end;
  FINALLY
    FreeAndNil(slModules);
  END;
end;

procedure TStandardExceptionInfo.GetJclDebugStackList;
begin
  { This is how you can retrieve the JclDebug Stack List information }
  with JclLastExceptStackList do
  begin
    ExceptionInfo.Add(Format(RsStackList, [DateTimeToStr(TimeStamp)]));
    AddToStrings(ExceptionInfo);
  end;
  ExceptionInfo.Add('');
end;

procedure TStandardExceptionInfo.GetOSInformation;
const
  OSNames: array [JclSysInfo.TWindowsVersion] of string =
    ('?', '95', '95OSR2', '98', '98SE', 'ME', 'NT3.1', 'NT3.5', 'NT3.51', 'NT4',
     '2000', 'XP');
  MMXText: array[Boolean] of PChar = ('', 'MMX');
  FDIVText: array[Boolean] of PChar = (' [FDIV Bug]', '');
var
  CpuInfo: TCpuInfo;
begin
  { Just some basic, but often handy-to-have OS information }
  ExceptionInfo.Add(Format(RsOSVersion, [OSNames[GetWindowsVersion],
    Win32MajorVersion, Win32MinorVersion, Win32BuildNumber, Win32CSDVersion]));
  GetCpuInfo(CpuInfo);
  with CpuInfo do
    ExceptionInfo.Add(Format(RsProcessor, [Manufacturer, CpuName,
      RoundFrequency(FrequencyInfo.NormFreq),
      MMXText[MMX], FDIVText[IsFDIVOK]]));
  ExceptionInfo.Add(Format(RsScreenRes, [GetDisplayWidth, GetDisplayHeight, Self.GetBPP]));
  ExceptionInfo.Add('');
end;

end.
